@@ -21,7 +21,7 @@
 */
 
 #include <config.h>
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include <string.h>
 #include <glade/glade.h>
 
@@ -244,7 +244,6 @@ selection_changed (GtkTreeSelection *selection, GtkHTMLEditTemplateProperties *d
 		else {
 			FILL;
 		}
-		CHANGE;
 	}
 }
 
@@ -278,7 +277,7 @@ template_widget (GtkHTMLEditTemplateProperties *d, gboolean insert)
 	gtk_tree_view_set_model (GTK_TREE_VIEW (d->tview_template), GTK_TREE_MODEL (d->store));
 
 	gtk_tree_view_append_column (GTK_TREE_VIEW (d->tview_template),
-				     gtk_tree_view_column_new_with_attributes (_("Labels"),
+				     gtk_tree_view_column_new_with_attributes (_("Template Labels"),
 									       gtk_cell_renderer_text_new (),
 									       "text", 0, NULL));
 	g_signal_connect (gtk_tree_view_get_selection (GTK_TREE_VIEW (d->tview_template)), "changed",
@@ -310,7 +309,6 @@ template_insert (GtkHTMLControlData *cd, gpointer *set_data)
 	*set_data = data;
 	rv = template_widget (data, TRUE);
 	set_ui (data);
-	gtk_html_edit_properties_dialog_change (data->cd->properties_dialog);
 
 	return rv;
 }
