@@ -23,6 +23,7 @@
 
 #include <config.h>
 #include "htmlclue.h"
+#include "htmlclueflow.h"
 #include "htmlobject.h"
 #include "htmlsearch.h"
 #include "htmltextslave.h"
@@ -244,8 +245,8 @@ split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint lev
 	*left  = g_list_prepend (*left, self);
 	*right = g_list_prepend (*right, dup);
 
-	level--;
-	if (level)
+	level --;
+	if (level > 0)
 		html_object_split (self->parent, e, dup, 0, level, left, right);
 }
 
@@ -422,8 +423,7 @@ check_point (HTMLObject *o,
 }
 
 static gint
-check_page_split (HTMLObject *self,
-		  gint y)
+check_page_split (HTMLObject *self, HTMLPainter *painter, gint y)
 {
 	HTMLClue *clue;
 	HTMLObject *p;
@@ -440,7 +440,7 @@ check_page_split (HTMLObject *self,
 			return last_under;
 
 		if (y >= y1 && y < y2)
-			return html_object_check_page_split (p, y - y1) + y1;
+			return html_object_check_page_split (p, painter, y - y1) + y1;
 		last_under = y2;
 	}
 
@@ -919,3 +919,15 @@ html_clue_remove_text_slaves (HTMLClue *clue)
 		}
 	}
 }
+
+
+gboolean
+html_clue_is_empty (HTMLClue *clue)
+{
+	if (clue->head == NULL)
+		return TRUE;
+	if (clue->head == clue->tail
+	    && HTML_IS_CLUEFLOW (clue->head) && html_clueflow_is_empty (HTML_CLUEFLOW (clue->head)))
+		return TRUE;
+	return FALSE;
+}
