@@ -25,6 +25,7 @@
 #include <string.h>
 #include <stdio.h>
 #include <gtk/gtksignal.h>
+#include <glib/gi18n.h>
 #include "gtkhtml.h"
 #include "gtkhtml-embedded.h"
 #include "htmlembedded.h"
@@ -92,7 +93,7 @@ draw (HTMLObject *o,
 		}
 	}
 
-	d (printf ("draw embedded %p - call painter\n", element));
+	d (printf ("draw embedded %p - call painter tx %d ty %d\n", element, tx, ty));
 	html_painter_draw_embedded (p, element, tx, ty);
 }
 
