@@ -1195,12 +1195,21 @@ insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir, gboolean add_undo)
 		html_undo_level_end (e->undo);
 
 	html_engine_thaw (e);
+	g_signal_emit_by_name (e->widget, "object_inserted", 0, 0);
 }
 
 void
 html_engine_insert_empty_paragraph (HTMLEngine *e)
 {
+	HTMLClueFlow *cf;
+
+	html_engine_freeze (e);
 	insert_empty_paragraph (e, HTML_UNDO_UNDO, TRUE);
+	cf = html_object_get_flow (e->cursor->object);
+	if (cf) {
+		cf->dir = html_text_direction_pango_to_html (gdk_keymap_get_direction (gdk_keymap_get_for_display (gtk_widget_get_display (GTK_WIDGET (e->widget)))));
+	}
+	html_engine_thaw (e);
 }
 
 static char *picto_chars = "DO)(|/PQ\0:-\0:\0:-\0:\0:;=-\0:;\0:-~\0:\0:\0:-\0:\0:-\0:\0:-\0:\0:-\0:\0";
