@@ -66,8 +66,14 @@ struct _HTMLObject {
            is set to TRUE instead of g_free()ing the object.  When the draw
            queue is flushed, the g_free() is performed.  */
 	guint free_pending : 1;
+	
+	/* FIXME add the other dynamic pusedo-classes... */
+	guint draw_focused : 1; 
 
 	GData *object_data;
+	GData *object_data_nocp;
+
+	char *id;
 };
 
 struct _HTMLObjectClearRectangle {
@@ -92,6 +98,7 @@ struct _HTMLObjectClass {
 
 	/* copy/cut/paste operations */
 	HTMLObject * (* op_copy)         (HTMLObject *self,
+					  HTMLObject *parent,
 					  HTMLEngine *e,
 					  GList      *from,
 					  GList      *to,
@@ -122,14 +129,13 @@ struct _HTMLObjectClass {
 	/* Layout management and geometry handling.  */
 
 	HTMLFitType (* fit_line) (HTMLObject *o, HTMLPainter *painter,
-				  gboolean start_of_line, gboolean first_run,
+				  gboolean start_of_line, gboolean first_run, gboolean next_to_floating,
 				  gint width_left);
 	gboolean (* calc_size) (HTMLObject *o, HTMLPainter *painter, GList **changed_objs);
 	gint (* calc_min_width) (HTMLObject *o, HTMLPainter *painter);
 	gint (* calc_preferred_width) (HTMLObject *o, HTMLPainter *painter);
-	void (* set_max_ascent) (HTMLObject *o, HTMLPainter *painter, gint a);
-	void (* set_max_descent) (HTMLObject *o, HTMLPainter *painter, gint d);
 	void (* set_max_width) (HTMLObject *o, HTMLPainter *painter, gint max_width);
+	void (* set_max_height) (HTMLObject *o, HTMLPainter *painter, gint max_height);
 
 	/* Relayout object `o' starting from child `child'.  This
            method can be called by the child when it changes any of
@@ -153,15 +159,14 @@ struct _HTMLObjectClass {
 
 	gboolean (* is_transparent) (HTMLObject *self);
 
-	/* This draws the background only.  If the object is
-           transparent, it should simply forward the method to the
-           parent.  */
-
+	/* Unused ::draw_background method */
 	void (* draw_background) (HTMLObject *o,
 				  HTMLPainter *painter,
 				  gint x, gint y,
 				  gint width, gint height,
 				  gint tx, gint ty);
+	/* remove it later */
+
 	void       (* set_bg_color) (HTMLObject *o, GdkColor *color);
 	GdkColor * (* get_bg_color) (HTMLObject *o, HTMLPainter *p);
 
@@ -170,8 +175,8 @@ struct _HTMLObjectClass {
            HTMLTable does not derive from HTMLClue and we don't want
            to spend time reorganizing the hierarchy now.  */
 
-	gint (* get_left_margin) (HTMLObject *self, HTMLPainter *painter, gint y);
-	gint (* get_right_margin) (HTMLObject *self, HTMLPainter *painter, gint y);
+	gint (* get_left_margin) (HTMLObject *self, HTMLPainter *painter, gint y, gboolean with_aligned);
+	gint (* get_right_margin) (HTMLObject *self, HTMLPainter *painter, gint y, gboolean with_aligned);
 
 	void (* set_painter) (HTMLObject *o, HTMLPainter *painter);
 
@@ -180,8 +185,8 @@ struct _HTMLObjectClass {
 
 	void (* reset) (HTMLObject *o);
 
-	const gchar * (* get_url)    (HTMLObject *o);
-	const gchar * (* get_target) (HTMLObject *o);
+	const gchar * (* get_url)    (HTMLObject *o, gint offset);
+	const gchar * (* get_target) (HTMLObject *o, gint offset);
 	const gchar * (* get_src)    (HTMLObject *o);
 
 	HTMLAnchor * (* find_anchor) (HTMLObject *o, const gchar *name, gint *x, gint *y);
@@ -224,7 +229,7 @@ struct _HTMLObjectClass {
 
 	/* Page splitting (for printing).  */
 
-	gint (* check_page_split) (HTMLObject *self, gint y);
+	gint (* check_page_split) (HTMLObject *self, HTMLPainter *p, gint y);
 
 	/* Selection.  */
 	gboolean (* select_range) (HTMLObject *self, HTMLEngine *engine, guint start, gint length,
@@ -248,12 +253,26 @@ struct _HTMLObjectClass {
 	HTMLObject * (* prev)            (HTMLObject *self, HTMLObject *child);
 	HTMLObject * (* head)            (HTMLObject *self);
 	HTMLObject * (* tail)            (HTMLObject *self);
+
+	HTMLClearType (* get_clear)      (HTMLObject *self);
+
+	gint (* get_n_children)          (HTMLObject *self);
+	HTMLObject * (* get_child)       (HTMLObject *self, gint index);
+	gint (* get_child_index)         (HTMLObject *self, HTMLObject *child);
+
+	HTMLDirection (*get_direction)   (HTMLObject *self);
+
+	gboolean (*cursor_forward)       (HTMLObject *self, HTMLCursor *cursor);
+	gboolean (*cursor_backward)       (HTMLObject *self, HTMLCursor *cursor);
+	gboolean (*cursor_right)         (HTMLObject *self, HTMLCursor *cursor);
+	gboolean (*cursor_left)          (HTMLObject *self, HTMLCursor *cursor);
+
+	int (*get_right_edge_offset) (HTMLObject *o, int offset);
+	int (*get_left_edge_offset) (HTMLObject *o, int offset);
 };
 
-
 extern HTMLObjectClass html_object_class;
 
-
 /* Basics.  */
 void            html_object_type_init             (void);
 void            html_object_init                  (HTMLObject            *self,
@@ -269,6 +288,7 @@ HTMLObject     *html_object_dup                   (HTMLObject            *self);
 
 /* copy/cut/paste operations */
 HTMLObject     *html_object_op_copy               (HTMLObject            *self,
+						   HTMLObject            *parent,
 						   HTMLEngine            *e,
 						   GList                 *from,
 						   GList                 *to,
@@ -299,18 +319,21 @@ void            html_object_set_parent            (HTMLObject            *self,
 						   HTMLObject            *parent);
 gint            html_object_get_left_margin       (HTMLObject            *self,
 						   HTMLPainter           *painter,
-						   gint                   y);
+						   gint                   y,
+						   gboolean               with_aligned);
 gint            html_object_get_right_margin      (HTMLObject            *self,
 						   HTMLPainter           *painter,
-						   gint                   y);
+						   gint                   y,
+						   gboolean               with_aligned);
 void            html_object_set_painter           (HTMLObject            *o,
 						   HTMLPainter           *p);
-void            html_object_clear_word_width      (HTMLObject            *o);
 void            html_object_reset                 (HTMLObject            *o);
 gboolean        html_object_is_text               (HTMLObject            *object);
 gboolean        html_object_is_clue               (HTMLObject            *object);
 HTMLEngine     *html_object_get_engine            (HTMLObject            *self,
 						   HTMLEngine            *e);
+HTMLEngine     *html_object_engine                (HTMLObject            *o,
+						   HTMLEngine            *e);
 void            html_object_forall                (HTMLObject            *self,
 						   HTMLEngine            *e,
 						   HTMLObjectForallFunc   func,
@@ -344,14 +367,6 @@ void            html_object_draw                  (HTMLObject            *o,
 						   gint                   height,
 						   gint                   tx,
 						   gint                   ty);
-void            html_object_draw_background       (HTMLObject            *o,
-						   HTMLPainter           *p,
-						   gint                   x,
-						   gint                   y,
-						   gint                   width,
-						   gint                   height,
-						   gint                   tx,
-						   gint                   ty);
 void            html_object_set_bg_color          (HTMLObject            *o,
 						   GdkColor              *color);
 GdkColor       *html_object_get_bg_color          (HTMLObject            *o,
@@ -363,19 +378,17 @@ HTMLFitType     html_object_fit_line              (HTMLObject            *o,
 						   HTMLPainter           *painter,
 						   gboolean               start_of_line,
 						   gboolean               first_run,
+						   gboolean               next_to_floating,
 						   gint                   width_left);
 gboolean        html_object_calc_size             (HTMLObject            *o,
 						   HTMLPainter           *painter,
 						   GList                **changed_objs);
-void            html_object_set_max_ascent        (HTMLObject            *o,
-						   HTMLPainter           *painter,
-						   gint                   a);
-void            html_object_set_max_descent       (HTMLObject            *o,
-						   HTMLPainter           *painter,
-						   gint                   d);
 void            html_object_set_max_width         (HTMLObject            *o,
 						   HTMLPainter           *painter,
 						   gint                   max_width);
+void            html_object_set_max_height        (HTMLObject            *o,
+						   HTMLPainter           *painter,
+						   gint                   max_height);
 gint            html_object_calc_min_width        (HTMLObject            *o,
 						   HTMLPainter           *painter);
 gint            html_object_calc_preferred_width  (HTMLObject            *o,
@@ -383,9 +396,12 @@ gint            html_object_calc_preferred_width  (HTMLObject            *o,
 void            html_object_calc_abs_position     (HTMLObject            *o,
 						   gint                  *x_return,
 						   gint                  *y_return);
-void            html_object_calc_intersection     (HTMLObject            *o,
-						   ArtIRect              *intersection,
-						   gint                   x,
+void            html_object_calc_abs_position_in_frame (HTMLObject       *o,
+							int              *x_return,
+							int              *y_return);
+gboolean        html_object_intersect             (HTMLObject            *o,
+					           GdkRectangle          *intersection,
+					           gint                   x,
 						   gint                   y,
 						   gint                   width,
 						   gint                   height);
@@ -395,8 +411,12 @@ gboolean        html_object_relayout              (HTMLObject            *obj,
 HTMLVAlignType  html_object_get_valign            (HTMLObject            *self);
 
 /* Links.  */
-const gchar    *html_object_get_url               (HTMLObject            *o);
-const gchar    *html_object_get_target            (HTMLObject            *o);
+const gchar    *html_object_get_url               (HTMLObject            *o,
+						   gint                   offset);
+const gchar    *html_object_get_target            (HTMLObject            *o,
+						   gint                   offset);
+gchar          *html_object_get_complete_url      (HTMLObject            *o,
+						   gint                   offset);
 const gchar    *html_object_get_src               (HTMLObject            *o);
 HTMLAnchor     *html_object_find_anchor           (HTMLObject            *o,
 						   const gchar           *name,
@@ -444,6 +464,10 @@ gboolean        html_object_cursor_forward        (HTMLObject            *self,
 						   HTMLCursor            *cursor);
 gboolean        html_object_cursor_backward       (HTMLObject            *self,
 						   HTMLCursor            *cursor);
+gboolean        html_object_cursor_left           (HTMLObject            *self,
+						   HTMLCursor            *cursor);
+gboolean        html_object_cursor_right          (HTMLObject            *self,
+						   HTMLCursor            *cursor);
 
 /* get prev/next object in scope of parent */
 HTMLObject     *html_object_next                  (HTMLObject            *self,
@@ -468,10 +492,13 @@ HTMLObject     *html_object_next_cursor           (HTMLObject            *self,
 						   gint                  *offset);
 HTMLObject     *html_object_prev_cursor           (HTMLObject            *self,
 						   gint                  *offset);
+GdkRectangle   *html_object_get_bounds            (HTMLObject            *o,
+						   GdkRectangle          *bounds);
 
 /* Page splitting.  */
-gint  html_object_check_page_split  (HTMLObject *self,
-				     gint        y);
+gint  html_object_check_page_split  (HTMLObject  *self,
+				     HTMLPainter *p,
+				     gint         y);
 
 /* Selection.  */
 gboolean    html_object_select_range             (HTMLObject *self,
@@ -502,13 +529,31 @@ void  html_object_change_set_down  (HTMLObject      *self,
 
 /* object data */
 
-void      html_object_set_data               (HTMLObject  *object,
-					      const gchar *key,
-					      const gchar *value);
-gpointer  html_object_get_data               (HTMLObject  *object,
-					      const gchar *key);
-void      html_object_copy_data_from_object  (HTMLObject  *dst,
-					      HTMLObject  *src);
+void      html_object_set_data               (HTMLObject          *object,
+					      const gchar         *key,
+					      const gchar         *value);
+void      html_object_set_data_full          (HTMLObject          *object,
+					      const gchar         *key,
+					      const gpointer       value,
+					      GDestroyNotify       func);
+gpointer  html_object_get_data               (HTMLObject          *object,
+					      const gchar         *key);
+void      html_object_copy_data_from_object  (HTMLObject          *dst,
+					      HTMLObject          *src);
+gboolean  html_object_save_data              (HTMLObject          *self,
+					      HTMLEngineSaveState *state);
+
+/* for acc object */
+void      html_object_set_data_nocp          (HTMLObject          *object,
+					      const gchar         *key,
+					      const gchar         *value);
+void      html_object_set_data_full_nocp     (HTMLObject          *object,
+					      const gchar         *key,
+					      const gpointer       value,
+					      GDestroyNotify       func);
+gpointer  html_object_get_data_nocp          (HTMLObject          *object,
+					      const gchar         *key);
+
 /*
  * editing
 */
@@ -550,4 +595,33 @@ gboolean  html_object_engine_intersection  (HTMLObject *o,
 void  html_object_add_to_changed  (GList      **changed_objs,
 				   HTMLObject  *o);
 
+gint html_object_get_n_children (HTMLObject *self);
+HTMLObject * html_object_get_child (HTMLObject *self, gint index);
+gint html_object_get_child_index (HTMLObject *self, HTMLObject *child);
+
+HTMLClearType html_object_get_clear (HTMLObject *self);
+
+HTMLObject *html_object_next_cursor_object  (HTMLObject *o,
+					     HTMLEngine *e,
+					     gint       *offset);
+HTMLObject *html_object_prev_cursor_object  (HTMLObject *o,
+					     HTMLEngine *e,
+					     gint       *offset);
+HTMLObject *html_object_next_cursor_leaf    (HTMLObject *o,
+					     HTMLEngine *e);
+HTMLObject *html_object_prev_cursor_leaf    (HTMLObject *o,
+					     HTMLEngine *e);
+
+int  html_object_get_right_edge_offset  (HTMLObject *o,
+					 int offset);
+int  html_object_get_left_edge_offset   (HTMLObject *o,
+					 int offset);
+
+const char *html_object_get_id  (HTMLObject *o);
+void        html_object_set_id  (HTMLObject *o,
+				 const char *id);
+
+HTMLDirection html_object_get_direction (HTMLObject *o);
+HTMLClueFlow * html_object_get_flow (HTMLObject *o);
+
 #endif /* _HTMLOBJECT_H_ */
