@@ -45,8 +45,6 @@
 HTMLTextSlaveClass html_text_slave_class;
 static HTMLObjectClass *parent_class = NULL;
 
-static GList * get_glyphs_part (HTMLTextSlave *slave, HTMLPainter *painter, guint offset, guint len);
-static void    clear_glyphs (HTMLTextSlave *slave);
 static void    clear_glyph_items (HTMLTextSlave *slave);
 
 char *
@@ -403,14 +401,26 @@ hts_fit_line (HTMLObject *o, HTMLPainter *painter,
 			if (update_lb (slave, painter, widthLeft, offset, s, ii, io, line_offset, w, &lwl, &lbw, &lbo, &lbsp, &force_fit))
 				break;
 
-		if (*s == '\t') {
+		if (io == 0 && slave->owner->text [pi->entries [ii].glyph_item.item->offset]  == '\t') {
+			GtkHTMLFontStyle font_style;
+			char *face;
 			gint skip = 8 - (line_offset % 8);
-			w += skip*pi->entries [ii].widths [io];
+
+			if (HTML_IS_PLAIN_PAINTER (painter)) {
+				font_style = GTK_HTML_FONT_STYLE_FIXED;
+				face = NULL;
+			} else {
+				font_style = html_text_get_font_style (slave->owner);
+				face = slave->owner->face;
+			}
+
+			pi->entries [ii].glyph_item.glyphs->glyphs[0].geometry.width = pi->entries [ii].widths [io]
+				= skip*html_painter_get_space_width (painter, font_style, face) * PANGO_SCALE;
 			line_offset += skip;
 		} else {
-			w += pi->entries [ii].widths [io];
 			line_offset ++;
 		}
+		w += pi->entries [ii].widths [io];
 
 		html_text_pi_forward (pi, &ii, &io);
 		s = g_utf8_next_char (s);
@@ -464,62 +474,6 @@ get_ys (HTMLText *text, HTMLPainter *p)
 		return 0;
 }
 
-static void
-draw_spell_errors (HTMLTextSlave *slave, HTMLPainter *p, gint tx, gint ty)
-{
-	GList *cur = HTML_TEXT (slave->owner)->spell_errors;
-	HTMLObject *obj = HTML_OBJECT (slave);
-	SpellError *se;
-	guint ma, mi;
-	gint x_off = 0;
-	gint last_off = 0;
-	gint line_offset = html_text_slave_get_line_offset (slave, 0, p);
-	gchar *text = html_text_slave_get_text (slave);
-	HTMLEngine *e;
-
-	if (p->widget && GTK_IS_HTML (p->widget))
-		e = GTK_HTML (p->widget)->engine;
-	else
-		return;
-
-	while (cur) {
-
-		se = (SpellError *) cur->data;
-		ma = MAX (se->off, slave->posStart);
-		mi = MIN (se->off + se->len, slave->posStart + slave->posLen);
-		if (ma < mi) {
-			GList *glyphs;
-			guint off = ma - slave->posStart;
-			guint len = mi - ma;
-			gint width, asc, dsc;
-
-			html_painter_set_pen (p, &html_colorset_get_color_allocated (e->settings->color_set,
-										     p, HTMLSpellErrorColor)->color);
-			/* printf ("spell error: %s\n", html_text_get_text (slave->owner, off)); */
-			
-			glyphs = get_glyphs_part (slave, p, last_off, off - last_off);
-			html_text_calc_text_size (slave->owner, p, text - slave->owner->text,
-						  off - last_off, html_text_get_pango_info (slave->owner, p), glyphs,
-						  &line_offset, &width, &asc, &dsc);
-			glyphs_destroy (glyphs);
-			x_off += width;
-			text = g_utf8_offset_to_pointer (text, off - last_off);
-			glyphs = get_glyphs_part (slave, p, off, len);
-			x_off += html_painter_draw_spell_error (p, obj->x + tx + x_off,
-								obj->y + ty + get_ys (slave->owner, p),
-								html_text_get_pango_info (slave->owner, p), glyphs);
-			glyphs_destroy (glyphs);
-			last_off = off + len;
-			if (line_offset != -1)
-				line_offset += len;
-			text = g_utf8_offset_to_pointer (text, len);
-		}
-		if (se->off > slave->posStart + slave->posLen)
-			break;
-		cur = cur->next;
-	}
-}
-
 static inline GList *
 get_glyphs_base_text (GList *glyphs, PangoItem *item, gint ii, const gchar *text, gint bytes)
 {
@@ -552,42 +506,6 @@ html_get_glyphs_non_tab (GList *glyphs, PangoItem *item, gint ii, const gchar *t
 	return glyphs;
 }
 
-static GList *
-get_glyphs_part (HTMLTextSlave *slave, HTMLPainter *painter, guint offset, guint len)
-{
-	GList *glyphs = NULL;
-	HTMLTextPangoInfo *pi;
-
-	pi = html_text_get_pango_info (slave->owner, painter);
-	if (pi) {
-		PangoItem *item;
-		gint index, c_len;
-		gint byte_offset, ii;
-		const gchar *text, *owner_text;
-		gchar *end;
-
-		owner_text = slave->owner->text;
-		text = g_utf8_offset_to_pointer (html_text_slave_get_text (slave), offset);
-		byte_offset = text - owner_text;
-
-		ii = html_text_pango_info_get_index (pi, byte_offset, 0);
-		index = 0;
-		while (index < len) {
-			item = pi->entries [ii].glyph_item.item;
-			c_len = MIN (item->num_chars - g_utf8_pointer_to_offset (owner_text + item->offset, text), len - index);
-
-			end = g_utf8_offset_to_pointer (text, c_len);
-			glyphs = html_get_glyphs_non_tab (glyphs, item, ii, text, end - text, c_len);
-			text = end;
-			index += c_len;
-			ii ++;
-		}
-		glyphs = g_list_reverse (glyphs);
-	}
-
-	return glyphs;
-}
-
 /*
  * NB: This implement the exact same algorithm as
  *     reorder-items.c:pango_reorder_items().
@@ -735,7 +653,7 @@ get_glyph_items_in_range (HTMLTextSlave *slave, HTMLPainter *painter, int start_
 GSList *
 html_text_slave_get_glyph_items (HTMLTextSlave *slave, HTMLPainter *painter)
 {
-	if (!slave->glyph_items || (HTML_OBJECT (slave)->change & HTML_CHANGE_RECALC_PI)) {
+	if (painter && (!slave->glyph_items || (HTML_OBJECT (slave)->change & HTML_CHANGE_RECALC_PI))) {
 		clear_glyph_items (slave);
 
 		HTML_OBJECT (slave)->change &= ~HTML_CHANGE_RECALC_PI;
@@ -745,6 +663,65 @@ html_text_slave_get_glyph_items (HTMLTextSlave *slave, HTMLPainter *painter)
 	return slave->glyph_items;
 }
 
+static gboolean
+calc_glyph_range_size (HTMLText *text, PangoGlyphItem *glyph_item, int start_index, int end_index, int *x_offset, int *width, int *asc, int *height)
+{
+	int isect_start, isect_end;
+
+	isect_start = MAX (glyph_item->item->offset, start_index);
+	isect_end = MIN (glyph_item->item->offset + glyph_item->item->length, end_index);
+
+	isect_start -= glyph_item->item->offset;
+	isect_end -= glyph_item->item->offset;
+
+	/* printf ("calc_glyph_range_size isect start %d end %d (end_index %d)\n", isect_start, isect_end, end_index); */
+
+	if (isect_start <= isect_end) {
+		PangoRectangle log_rect;
+		int start_x, end_x;
+		
+		pango_glyph_string_index_to_x (glyph_item->glyphs,
+					       text->text + glyph_item->item->offset,
+					       glyph_item->item->length,
+					       &glyph_item->item->analysis,
+					       isect_start,
+					       FALSE, &start_x);
+
+		if (isect_start < isect_end)
+			pango_glyph_string_index_to_x (glyph_item->glyphs,
+						       text->text + glyph_item->item->offset,
+						       glyph_item->item->length,
+						       &glyph_item->item->analysis,
+						       isect_end,
+						       FALSE, &end_x);
+		else
+			end_x = start_x;
+
+		if (asc || height)
+			/* this call is used only to get ascent and height */
+			pango_glyph_string_extents (glyph_item->glyphs, glyph_item->item->analysis.font, NULL, &log_rect);
+
+		/* printf ("selection_start_index %d selection_end_index %d isect_start %d isect_end %d start_x %d end_x %d cwidth %d width %d\n",
+		   selection_start_index, selection_end_index, isect_start, isect_end,
+		   html_painter_pango_to_engine (p, start_x), html_painter_pango_to_engine (p, end_x),
+		   html_painter_pango_to_engine (p, start_x < end_x ? (end_x - start_x) : (start_x - end_x)),
+		   html_painter_pango_to_engine (p, log_rect.width)); */
+
+		if (x_offset)
+			*x_offset = MIN (start_x, end_x);
+		if (width)
+			*width = start_x < end_x ? (end_x - start_x) : (start_x - end_x);
+		if (asc)
+			*asc = PANGO_ASCENT (log_rect);
+		if (height)
+			*height = log_rect.height;
+
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
 static void
 draw_text (HTMLTextSlave *self,
 	   HTMLPainter *p,
@@ -761,6 +738,7 @@ draw_text (HTMLTextSlave *self,
 	int isect_start, isect_end;
 	gboolean selection;
 	GdkColor selection_fg, selection_bg;
+	HTMLEngine *e = NULL;
 
 	obj = HTML_OBJECT (self);
 
@@ -768,8 +746,10 @@ draw_text (HTMLTextSlave *self,
 	isect_end = MIN (text->select_start + text->select_length, self->posStart + self->posLen);
 	selection = isect_start < isect_end;
 
+	if (p->widget && GTK_IS_HTML (p->widget))
+		e = GTK_HTML (p->widget)->engine;
+
 	if (selection) {
-		HTMLEngine *e = NULL;
 		gchar *end;
 		gchar *start;
 
@@ -779,9 +759,6 @@ draw_text (HTMLTextSlave *self,
 		selection_start_index = start - text->text;
 		selection_end_index = end - text->text;
 
-		if (p->widget && GTK_IS_HTML (p->widget))
-			e = GTK_HTML (p->widget)->engine;
-
 		if (e) {
 			selection_fg = html_colorset_get_color_allocated
 				(e->settings->color_set, p,
@@ -797,60 +774,68 @@ draw_text (HTMLTextSlave *self,
 	run_width = 0;
 	for (cur = html_text_slave_get_glyph_items (self, p); cur; cur = cur->next) {
 		HTMLTextSlaveGlyphItem *gi = (HTMLTextSlaveGlyphItem *) cur->data;
+		GList *cur_se;
 		int cur_width;
 
-		cur_width = html_painter_draw_glyphs (p, obj->x + tx + run_width, obj->y + ty + get_ys (text, p), gi->glyph_item.item, gi->glyph_item.glyphs, NULL, NULL);
+		cur_width = html_painter_draw_glyphs (p, obj->x + tx + html_painter_pango_to_engine (p, run_width),
+						      obj->y + ty + get_ys (text, p), gi->glyph_item.item, gi->glyph_item.glyphs, NULL, NULL);
 
 		if (selection) {
-			isect_start = MAX (gi->glyph_item.item->offset, selection_start_index);
-			isect_end = MIN (gi->glyph_item.item->offset + gi->glyph_item.item->length, selection_end_index);
-
-			isect_start -= gi->glyph_item.item->offset;
-			isect_end -= gi->glyph_item.item->offset;
-
-			if (isect_start < isect_end) {
-				PangoRectangle log_rect;
-				int start_x, end_x;
-				int cx, cy, cw, ch;
-		
-				pango_glyph_string_index_to_x (gi->glyph_item.glyphs,
-							       text->text + gi->glyph_item.item->offset,
-							       gi->glyph_item.item->length,
-							       &gi->glyph_item.item->analysis,
-							       isect_start,
-							       FALSE, &start_x);
-
-				pango_glyph_string_index_to_x (gi->glyph_item.glyphs,
-							       text->text + gi->glyph_item.item->offset,
-							       gi->glyph_item.item->length,
-							       &gi->glyph_item.item->analysis,
-							       isect_end,
-							       FALSE, &end_x);
-
-				/* this call is used only to get ascent and height */
-				pango_glyph_string_extents (gi->glyph_item.glyphs, gi->glyph_item.item->analysis.font, NULL, &log_rect);
-
-				/* printf ("selection_start_index %d selection_end_index %d isect_start %d isect_end %d start_x %d end_x %d cwidth %d width %d\n",
-					selection_start_index, selection_end_index, isect_start, isect_end,
-					html_painter_pango_to_engine (p, start_x), html_painter_pango_to_engine (p, end_x),
-					html_painter_pango_to_engine (p, start_x < end_x ? (end_x - start_x) : (start_x - end_x)),
-					html_painter_pango_to_engine (p, log_rect.width)); */
+			int start_x, width, asc, height;
+			int cx, cy, cw, ch;
 
+			if (calc_glyph_range_size (text, &gi->glyph_item, selection_start_index, selection_end_index, &start_x, &width, &asc, &height) && width > 0) {
 				html_painter_get_clip_rectangle (p, &cx, &cy, &cw, &ch);
+/* 				printf ("run_width; %d start_x %d index %d\n", run_width, start_x, selection_start_index); */
 				html_painter_set_clip_rectangle (p,
-							obj->x + tx + run_width + html_painter_pango_to_engine (p, MIN (start_x, end_x)),
-							obj->y + ty + get_ys (text, p) - html_painter_pango_to_engine (p, PANGO_ASCENT (log_rect)),
-							html_painter_pango_to_engine (p, start_x < end_x ? (end_x - start_x) : (start_x - end_x)),
-							html_painter_pango_to_engine (p, log_rect.height));
+							obj->x + tx + html_painter_pango_to_engine (p, run_width + start_x),
+							obj->y + ty + get_ys (text, p) - html_painter_pango_to_engine (p, asc),
+							html_painter_pango_to_engine (p, width),
+							html_painter_pango_to_engine (p, height));
 
 				/* printf ("draw selection %d %d %d at %d, %d\n", selection_bg.red, selection_bg.green, selection_bg.blue,
 				   obj->x + tx + run_width, obj->y + ty + get_ys (text, p)); */
-				html_painter_draw_glyphs (p, obj->x + tx + run_width, obj->y + ty + get_ys (text, p), gi->glyph_item.item, gi->glyph_item.glyphs,
+				html_painter_draw_glyphs (p, obj->x + tx + html_painter_pango_to_engine (p, run_width),
+							  obj->y + ty + get_ys (text, p), gi->glyph_item.item, gi->glyph_item.glyphs,
 							  &selection_fg, &selection_bg);
 				html_painter_set_clip_rectangle (p, cx, cy, cw, ch);
 			}
 		}
 
+		for (cur_se = text->spell_errors; e && cur_se; cur_se = cur_se->next) {
+			SpellError *se;
+			guint ma, mi;
+
+			se = (SpellError *) cur_se->data;
+			ma = MAX (se->off, self->posStart);
+			mi = MIN (se->off + se->len, self->posStart + self->posLen);
+
+			if (ma < mi) {
+				int width, height, asc, start_x;
+
+				gchar *end;
+				gchar *start;
+				int se_start_index, se_end_index;
+
+				start = html_text_get_text (text, ma);
+				end = g_utf8_offset_to_pointer (start, mi - ma);
+
+				se_start_index = start - text->text;
+				se_end_index = end - text->text;
+
+				if (calc_glyph_range_size (text, &gi->glyph_item, se_start_index, se_end_index, &start_x, &width, &asc, &height)) {
+					html_painter_set_pen (p, &html_colorset_get_color_allocated (e->settings->color_set,
+												     p, HTMLSpellErrorColor)->color);
+				        /* printf ("spell error: %s\n", html_text_get_text (slave->owner, off)); */
+			
+					html_painter_draw_spell_error (p, obj->x + tx + html_painter_pango_to_engine (p, run_width + start_x),
+								       obj->y + ty + get_ys (self->owner, p), html_painter_pango_to_engine (p, width));
+				}
+			}
+			if (se->off > self->posStart + self->posLen)
+				break;
+		}
+
 		run_width += cur_width;
 	}
 }
@@ -932,9 +917,6 @@ draw (HTMLObject *o,
 
 	end = slave->posStart + slave->posLen;
 	draw_text (slave, p, font_style, x, y, width, height, tx, ty);
-
-	if (owner->spell_errors)
-		draw_spell_errors (slave, p, tx ,ty);
 	
 	if (HTML_OBJECT (owner)->draw_focused)
 		draw_focus (slave, p, tx, ty);
@@ -1013,7 +995,7 @@ calc_offset (HTMLTextSlave *slave, HTMLPainter *painter, gint x)
 		}
 	}
 
-	/* printf ("offset %d\n", offset); */
+/* 	printf ("offset %d\n", offset); */
 
 	return offset;
 }
@@ -1161,3 +1143,387 @@ html_text_slave_new (HTMLText *owner, guint posStart, guint posLen)
 
 	return HTML_OBJECT (slave);
 }
+
+static gboolean
+html_text_slave_is_index_in_glyph (HTMLTextSlave *slave, HTMLTextSlave *next_slave, GSList *cur, int index, PangoItem *item)
+{
+	if (item->analysis.level % 2 == 0) {
+		/* LTR */
+		return item->offset <= index
+			&& (index < item->offset + item->length
+			    || (index == item->offset + item->length &&
+				(!cur->next
+				 || (!next_slave && slave->owner->text_bytes == item->offset + item->length)
+				 || (next_slave && html_text_slave_get_text (next_slave) - next_slave->owner->text == item->offset + item->length))));
+	} else {
+		/* RTL */
+		return index <= item->offset + item->length
+			&& (item->offset < index
+			    || (index == item->offset &&
+				(!cur->next
+				 || (!next_slave && slave->owner->text_bytes == item->offset + item->length)
+				 || (next_slave && html_text_slave_get_text (next_slave) - next_slave->owner->text == item->offset))));
+	}
+}
+
+static HTMLTextSlaveGlyphItem *
+html_text_slave_get_glyph_item_at_offset (HTMLTextSlave *slave, HTMLPainter *painter, int offset, HTMLTextSlaveGlyphItem **prev, HTMLTextSlaveGlyphItem **next, int *start_width, int *index_out)
+{
+	HTMLTextSlaveGlyphItem *rv = NULL;
+	HTMLTextSlaveGlyphItem *prev_gi, *next_gi;
+	HTMLTextSlave *next_slave = HTML_OBJECT (slave)->next && HTML_IS_TEXT_SLAVE (HTML_OBJECT (slave)->next) ? HTML_TEXT_SLAVE (HTML_OBJECT (slave)->next) : NULL;
+	GSList *cur;
+	int index;
+
+	index = g_utf8_offset_to_pointer (html_text_slave_get_text (slave), offset) - slave->owner->text;
+	if (index_out)
+		*index_out = index;
+
+	if (start_width)
+		*start_width = 0;
+
+	cur = html_text_slave_get_glyph_items (slave, painter);
+	if (cur) {
+		for (prev_gi = NULL; cur; cur = cur->next) {
+			HTMLTextSlaveGlyphItem *gi = (HTMLTextSlaveGlyphItem *) cur->data;
+
+			if (html_text_slave_is_index_in_glyph (slave, next_slave, cur, index, gi->glyph_item.item)) {
+				next_gi = cur->next ? (HTMLTextSlaveGlyphItem *) cur->next->data : NULL;
+				rv = gi;
+				break;
+			}
+
+			prev_gi = gi;
+			if (start_width) {
+				PangoRectangle log_rect;
+
+				pango_glyph_string_extents (gi->glyph_item.glyphs, gi->glyph_item.item->analysis.font, NULL, &log_rect);
+				(*start_width) += log_rect.width;
+			}
+		}
+	} else {
+		prev_gi = next_gi = NULL;
+	}
+
+	if (prev)
+		*prev = prev_gi;
+
+	if (next)
+		*next = next_gi;
+
+	return rv;
+}
+
+static gboolean
+html_text_slave_gi_left_edge (HTMLTextSlave *slave, HTMLCursor *cursor, HTMLTextSlaveGlyphItem *gi)
+{
+	int old_offset = cursor->offset;
+
+	if (gi->glyph_item.item->analysis.level % 2 == 0) {
+		/* LTR */
+		cursor->offset = slave->posStart + g_utf8_pointer_to_offset (html_text_slave_get_text (slave),
+									     slave->owner->text + gi->glyph_item.item->offset);
+		cursor->position += cursor->offset - old_offset;
+	} else {
+		/* RTL */
+		cursor->offset = slave->posStart + g_utf8_pointer_to_offset (html_text_slave_get_text (slave),
+									     slave->owner->text + gi->glyph_item.item->offset + gi->glyph_item.item->length);
+		cursor->position += cursor->offset - old_offset;
+	}
+
+	return TRUE;
+}
+
+static gboolean
+html_text_slave_gi_right_edge (HTMLTextSlave *slave, HTMLCursor *cursor, HTMLTextSlaveGlyphItem *gi)
+{
+	int old_offset = cursor->offset;
+
+	if (gi->glyph_item.item->analysis.level % 2 == 0) {
+		/* LTR */
+		cursor->offset = slave->posStart + g_utf8_pointer_to_offset (html_text_slave_get_text (slave),
+									     slave->owner->text + gi->glyph_item.item->offset + gi->glyph_item.item->length);
+		cursor->position += cursor->offset - old_offset;
+	} else {
+		/* RTL */
+		cursor->offset = slave->posStart + g_utf8_pointer_to_offset (html_text_slave_get_text (slave),
+									     slave->owner->text + gi->glyph_item.item->offset);
+		cursor->position += cursor->offset - old_offset;
+	}
+
+	return TRUE;
+}
+
+static gboolean
+html_text_slave_cursor_right_one (HTMLTextSlave *slave, HTMLCursor *cursor)
+{
+	HTMLTextSlaveGlyphItem *prev, *next;
+	int index;
+	HTMLTextSlaveGlyphItem *gi = html_text_slave_get_glyph_item_at_offset (slave, NULL, cursor->offset - slave->posStart, &prev, &next, NULL, &index);
+
+	if (!gi)
+		return FALSE;
+
+	if (gi->glyph_item.item->analysis.level % 2 == 0) {
+		/* LTR */
+		if (index < gi->glyph_item.item->offset + gi->glyph_item.item->length) {
+			cursor->offset ++;
+			cursor->position ++;
+
+			return TRUE;
+		}
+	} else {
+		/* RTL */
+		if (index > gi->glyph_item.item->offset && (index <= gi->glyph_item.item->offset + gi->glyph_item.item->length)) {
+			cursor->offset --;
+			cursor->position --;
+
+			return TRUE;
+		}
+	}
+
+	if (next) {
+		if (html_text_slave_gi_left_edge (slave, cursor, next)) {
+			if (next->glyph_item.item->analysis.level % 2 == 0) {
+				/* LTR */
+				cursor->offset ++;
+				cursor->position ++;
+			} else {
+				/* RTL */
+				cursor->offset --;
+				cursor->position --;
+			}
+
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+gboolean
+html_text_slave_cursor_right (HTMLTextSlave *slave, HTMLCursor *cursor)
+{
+	HTMLTextPangoInfo *pi = html_text_get_pango_info (slave->owner, NULL);
+	gboolean step_success;
+
+	do
+		step_success = html_text_slave_cursor_right_one (slave, cursor);
+	while (step_success && !pi->attrs [cursor->offset].is_cursor_position);
+
+	return step_success;
+}
+
+static gboolean
+html_text_slave_cursor_left_one (HTMLTextSlave *slave, HTMLCursor *cursor)
+{
+	HTMLTextSlaveGlyphItem *prev, *next;
+	int index;
+	HTMLTextSlaveGlyphItem *gi = html_text_slave_get_glyph_item_at_offset (slave, NULL, cursor->offset - slave->posStart, &prev, &next, NULL, &index);
+
+/* 	printf ("gi: %p item num chars: %d\n", gi, gi ? gi->glyph_item.item->num_chars : -1); */
+
+	if (!gi)
+		return FALSE;
+
+	if (gi->glyph_item.item->analysis.level % 2 == 0) {
+		/* LTR */
+		if (index - gi->glyph_item.item->offset > 1 || (!prev && index - gi->glyph_item.item->offset > 0)) {
+			cursor->offset --;
+			cursor->position --;
+
+			return TRUE;
+		}
+	} else {
+		/* RTL */
+		if (index < gi->glyph_item.item->offset + gi->glyph_item.item->length) {
+			cursor->offset ++;
+			cursor->position ++;
+
+			return TRUE;
+		}
+	}
+
+	if (prev) {
+		if (html_text_slave_gi_right_edge (slave, cursor, prev)) {
+			if (prev->glyph_item.item->analysis.level % 2 == 0) {
+				/* LTR */
+				if (index - gi->glyph_item.item->offset == 0) {
+					cursor->offset --;
+					cursor->position --;
+				}
+			} else {
+				/* RTL */
+				cursor->offset ++;
+				cursor->position ++;
+			}
+
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+gboolean
+html_text_slave_cursor_left (HTMLTextSlave *slave, HTMLCursor *cursor)
+{
+	HTMLTextPangoInfo *pi = html_text_get_pango_info (slave->owner, NULL);
+	gboolean step_success;
+
+	do
+		step_success = html_text_slave_cursor_left_one (slave, cursor);
+	while (step_success && !pi->attrs [cursor->offset].is_cursor_position);
+
+	return step_success;
+}
+
+static gboolean
+html_text_slave_get_left_edge (HTMLTextSlave *slave, HTMLCursor *cursor)
+{
+	HTMLTextPangoInfo *pi = html_text_get_pango_info (slave->owner, NULL);
+	int old_offset = cursor->offset;
+	int old_position = cursor->position;
+
+	cursor->offset = html_text_slave_get_left_edge_offset (slave);
+
+	if (pi->attrs [cursor->offset].is_cursor_position && old_offset != cursor->offset)
+		return TRUE;
+	else {
+		if (html_text_slave_cursor_right (slave, cursor)) {
+			/* we should preserve position here as caller function correct position themselves */
+			cursor->position = old_position;
+			return TRUE;
+		} else
+			return FALSE;
+	}
+}
+
+static gboolean
+html_text_slave_get_right_edge (HTMLTextSlave *slave, HTMLCursor *cursor)
+{
+	HTMLTextPangoInfo *pi = html_text_get_pango_info (slave->owner, NULL);
+	int old_offset = cursor->offset;
+	int old_position = cursor->position;
+
+	cursor->offset = html_text_slave_get_right_edge_offset (slave);
+
+	if (pi->attrs [cursor->offset].is_cursor_position && old_offset != cursor->offset)
+		return TRUE;
+	else {
+		if (html_text_slave_cursor_left (slave, cursor)) {
+			/* we should preserve position here as caller function correct position themselves */
+			cursor->position = old_position;
+			return TRUE;
+		} else
+			return FALSE;
+	}
+}
+
+gboolean
+html_text_slave_cursor_head (HTMLTextSlave *slave, HTMLCursor *cursor)
+{
+	if (html_text_slave_get_glyph_items (slave, NULL)) {
+		cursor->object = HTML_OBJECT (slave->owner);
+
+		if (html_text_get_pango_direction (slave->owner) != PANGO_DIRECTION_RTL) {
+			/* LTR */
+			return html_text_slave_get_left_edge (slave, cursor);
+		} else {
+			/* RTL */
+			return html_text_slave_get_right_edge (slave, cursor);
+		}
+	}
+
+	return FALSE;
+}
+
+gboolean
+html_text_slave_cursor_tail (HTMLTextSlave *slave, HTMLCursor *cursor)
+{
+	if (html_text_slave_get_glyph_items (slave, NULL)) {
+		cursor->object = HTML_OBJECT (slave->owner);
+
+		if (html_text_get_pango_direction (slave->owner) != PANGO_DIRECTION_RTL) {
+			/* LTR */
+			return html_text_slave_get_right_edge (slave, cursor);
+		} else {
+			/* RTL */
+			return html_text_slave_get_left_edge (slave, cursor);
+		}
+	}
+
+	return FALSE;
+}
+
+void
+html_text_slave_get_cursor_base (HTMLTextSlave *slave, HTMLPainter *painter, guint offset, gint *x, gint *y)
+{
+	HTMLTextSlaveGlyphItem *gi;
+	int index, start_width;
+
+	html_object_calc_abs_position (HTML_OBJECT (slave), x, y);
+
+	gi = html_text_slave_get_glyph_item_at_offset (slave, painter, (int) offset, NULL, NULL, &start_width, &index);
+
+/* 	printf ("gi: %p index: %d start_width: %d item indexes %d %d\n", */
+/* 		gi, index, start_width, gi ? gi->glyph_item.item->offset : -1, */
+/* 		gi ? gi->glyph_item.item->offset + gi->glyph_item.item->length : -1); */
+
+	if (gi) {
+		int start_x;
+
+		if (calc_glyph_range_size (slave->owner, &gi->glyph_item, index, index, &start_x, NULL, NULL, NULL) && x) {
+/* 			printf ("start_width: %d start_x: %d\n", start_width, start_x); */
+			*x += html_painter_pango_to_engine (painter, start_width + start_x);
+		}
+	}
+}
+
+int
+html_text_slave_get_left_edge_offset (HTMLTextSlave *slave)
+{
+	GSList *gis = html_text_slave_get_glyph_items (slave, NULL);
+
+	if (gis) {
+		HTMLTextSlaveGlyphItem *gi = (HTMLTextSlaveGlyphItem *) gis->data;
+
+		if (gi->glyph_item.item->analysis.level % 2 == 0) {
+			/* LTR */
+			return slave->posStart + g_utf8_pointer_to_offset (html_text_slave_get_text (slave), slave->owner->text + gi->glyph_item.item->offset);
+		} else {
+			/* RTL */
+			return slave->posStart + g_utf8_pointer_to_offset (html_text_slave_get_text (slave),
+									   slave->owner->text + gi->glyph_item.item->offset + gi->glyph_item.item->length);
+		}
+	} else {
+		if (slave->owner->text_len > 0)
+			g_warning ("html_text_slave_get_left_edge_offset failed");
+
+		return 0;
+	}
+}
+
+int
+html_text_slave_get_right_edge_offset (HTMLTextSlave *slave)
+{
+	GSList *gis = html_text_slave_get_glyph_items (slave, NULL);
+
+	if (gis) {
+		HTMLTextSlaveGlyphItem *gi = (HTMLTextSlaveGlyphItem *) g_slist_last (gis)->data;
+
+		if (gi->glyph_item.item->analysis.level % 2 == 0) {
+			/* LTR */
+			return slave->posStart + g_utf8_pointer_to_offset (html_text_slave_get_text (slave),
+									   slave->owner->text + gi->glyph_item.item->offset + gi->glyph_item.item->length);
+		} else {
+			/* RTL */
+			return slave->posStart + g_utf8_pointer_to_offset (html_text_slave_get_text (slave), slave->owner->text + gi->glyph_item.item->offset);
+		}
+	} else {
+		if (slave->owner->text_len > 0)
+			g_warning ("html_text_slave_get_left_edge_offset failed");
+
+		return 0;
+	}
+}
