@@ -1,5 +1,6 @@
 #include <string.h>
 #include <stdio.h>
+#include <glib/gstring.h>
 #include <gtk/gtkmain.h>
 #include <gtk/gtkwindow.h>
 #include "gtkhtml.h"
@@ -8,8 +9,14 @@
 #include "htmlcluev.h"
 #include "htmlcursor.h"
 #include "htmlengine.h"
+#include "htmlengine-edit.h"
+#include "htmlengine-edit-cut-and-paste.h"
 #include "htmlengine-edit-movement.h"
 #include "htmlengine-edit-text.h"
+#include "htmlengine-save.h"
+#include "htmlselection.h"
+#include "htmltable.h"
+#include "htmltablecell.h"
 #include "htmltext.h"
 
 typedef struct {
@@ -22,20 +29,32 @@ static int test_cursor_beol_rtl (GtkHTML *html);
 static int test_cursor_left_right_on_items_boundaries (GtkHTML *html);
 static int test_cursor_left_right_on_lines_boundaries (GtkHTML *html);
 static int test_cursor_left_right_on_lines_boundaries_rtl (GtkHTML *html);
+static int test_cursor_left_right_on_lines_boundaries_wo_white (GtkHTML *html);
+static int test_cursor_around_containers (GtkHTML *html);
 
 static int test_quotes_in_div_block (GtkHTML *html);
+static int test_quotes_in_table (GtkHTML *html);
 static int test_capitalize_upcase_lowcase_word (GtkHTML *html);
+static int test_delete_nested_cluevs_and_undo (GtkHTML *html);
+static int test_indentation_plain_text (GtkHTML *html);
+static int test_indentation_plain_text_rtl (GtkHTML *html);
 
 static Test tests[] = {
 	{ "cursor movement", NULL },
 	{ "left/right on items boundaries", test_cursor_left_right_on_items_boundaries },
 	{ "left/right on lines boundaries", test_cursor_left_right_on_lines_boundaries },
 	{ "left/right on lines boundaries (RTL)", test_cursor_left_right_on_lines_boundaries_rtl },
+	{ "left/right on line boundaries - without white space", test_cursor_left_right_on_lines_boundaries_wo_white },
 	{ "begin/end of line", test_cursor_beol },
 	{ "begin/end of line (RTL)", test_cursor_beol_rtl },
+	{ "around containers", test_cursor_around_containers },
 	{ "various fixed bugs", NULL },
 	{ "outer quotes inside div block", test_quotes_in_div_block },
+	{ "outer quotes inside table", test_quotes_in_table },
 	{ "capitalize, upcase/lowcase word", test_capitalize_upcase_lowcase_word },
+	{ "delete across nested cluev's and undo", test_delete_nested_cluevs_and_undo },
+	{ "indentation in plain text", test_indentation_plain_text },
+	{ "indentation in plain text (RTL)", test_indentation_plain_text_rtl },
 	{ NULL, NULL }
 };
 
@@ -47,6 +66,131 @@ static void load_editable (GtkHTML *html, char *s)
 	gtk_html_set_editable (html, TRUE);
 }
 
+static gboolean
+plain_save_receiver (gpointer engine, const char *data, unsigned int len, gpointer user_data)
+{
+	GString *str = (GString *) user_data;
+
+	g_string_append_len (str, data, len);
+
+	return TRUE;
+}
+
+static char *
+get_plain (GtkHTML *html)
+{
+	GString *str = g_string_new (0);
+	char *rv;
+
+	html_engine_save_plain (html->engine, plain_save_receiver, str);
+
+	rv = str->str;
+	g_string_free (str, FALSE);
+
+	return rv;
+}
+
+static int test_delete_nested_cluevs_and_undo (GtkHTML *html)
+{
+	load_editable (html, "<div>abc</div><div>efg</div>");
+
+	html_cursor_jump_to_position (html->engine->cursor, html->engine, 1);
+	if (html->engine->cursor->offset != 1
+	    || html->engine->cursor->position != 1)
+		return FALSE;
+
+	html_engine_set_mark (html->engine);
+	html_cursor_jump_to_position (html->engine->cursor, html->engine, 6);
+	if (html->engine->cursor->offset != 2
+	    || html->engine->cursor->position != 6)
+		return FALSE;
+
+	html_engine_delete (html->engine);
+
+	html_engine_end_of_document (html->engine);
+	if (html->engine->cursor->offset != 2
+	    || html->engine->cursor->position != 2)
+		return FALSE;
+
+	html_engine_undo (html->engine);
+
+	html_engine_end_of_document (html->engine);
+	if (html->engine->cursor->offset != 3
+	    || html->engine->cursor->position != 7)
+		return FALSE;
+
+	return TRUE;
+}
+
+static int test_indentation_plain_text (GtkHTML *html)
+{
+	char *str;
+
+	load_editable (html, "abc<div align=right>abc</div>");
+
+	str = get_plain (html);
+	if (!str || strcmp (str, "abc\n                                                                     abc\n"))
+		return FALSE;
+
+	g_free (str);
+
+	return TRUE;
+}
+
+static int test_indentation_plain_text_rtl (GtkHTML *html)
+{
+	char *str;
+
+	load_editable (html, "שנבגקכעי<div align=left>שנבגקכעי</div>");
+
+	str = get_plain (html);
+	if (!str || strcmp (str, "שנבגקכעי\n                                                                שנבגקכעי\n"))
+		return FALSE;
+
+	g_free (str);
+
+	return TRUE;
+}
+
+static int test_cursor_around_containers (GtkHTML *html)
+{
+	load_editable (html, "abc<table><tr><td>abc</td></tr></table>");
+
+	html_cursor_jump_to_position (html->engine->cursor, html->engine, 9);
+	if (html->engine->cursor->offset != 1
+	    || html->engine->cursor->position != 9
+	    || !html_cursor_beginning_of_line (html->engine->cursor, html->engine)
+	    || html->engine->cursor->offset != 0
+	    || html->engine->cursor->position != 4)
+		return FALSE;
+
+	html_cursor_jump_to_position (html->engine->cursor, html->engine, 4);
+	if (html->engine->cursor->offset != 0
+	    || html->engine->cursor->position != 4
+	    || !html_cursor_end_of_line (html->engine->cursor, html->engine)
+	    || html->engine->cursor->offset != 1
+	    || html->engine->cursor->position != 9)
+		return FALSE;
+
+	html_cursor_jump_to_position (html->engine->cursor, html->engine, 5);
+	if (html->engine->cursor->offset != 0
+	    || html->engine->cursor->position != 5
+	    || !html_cursor_left (html->engine->cursor, html->engine)
+	    || html->engine->cursor->offset != 0
+	    || html->engine->cursor->position != 4)
+		return FALSE;
+
+	html_cursor_jump_to_position (html->engine->cursor, html->engine, 8);
+	if (html->engine->cursor->offset != 3
+	    || html->engine->cursor->position != 8
+	    || !html_cursor_right (html->engine->cursor, html->engine)
+	    || html->engine->cursor->offset != 1
+	    || html->engine->cursor->position != 9)
+		return FALSE;
+
+	return TRUE;
+}
+
 static int test_cursor_left_right_on_items_boundaries (GtkHTML *html)
 {
 	load_editable (html, "ab<b>cde</b>ef");
@@ -163,6 +307,27 @@ static int test_cursor_left_right_on_lines_boundaries_rtl (GtkHTML *html)
 	return TRUE;
 }
 
+static int test_cursor_left_right_on_lines_boundaries_wo_white (GtkHTML *html)
+{
+	load_editable (html, "oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo%next line");
+
+	html_cursor_jump_to_position (html->engine->cursor, html->engine, 81);
+
+	if (html->engine->cursor->offset != 81
+	    || html->engine->cursor->position != 81
+	    || !html_cursor_left (html->engine->cursor, html->engine)
+	    || html->engine->cursor->offset != 80
+	    || html->engine->cursor->position != 80
+	    || !html_cursor_right (html->engine->cursor, html->engine)
+	    || html->engine->cursor->offset != 81
+	    || html->engine->cursor->position != 81)
+		return FALSE;
+
+	printf ("test_cursor_left_right_on_lines_boundaries_wo_white: passed\n");
+
+	return TRUE;
+}
+
 static int test_cursor_beol (GtkHTML *html)
 {
 	load_editable (html, "<pre>simple line\nsecond line\n");
@@ -204,8 +369,7 @@ static int test_cursor_beol (GtkHTML *html)
 
 static int test_cursor_beol_rtl (GtkHTML *html)
 {
-	load_editable (html, "<pre>أوروبا, برمجيات الحاسوب + انترنيت :\noooooooooooooooooooooooooooooooooooooooooooooooo");
-	//load_editable (html, "<pre>أوروبا, برمجيات الحاسوب + انترنيت :\nتصبح عالميا مع يونيكود\n");
+	load_editable (html, "<pre>أوروبا, برمجيات الحاسوب + انترنيت :\nتصبح عالميا مع يونيكود\n");
 
 	/* first test it on 1st line */
 	printf ("test_cursor_beol_rtl: testing 1st line eol\n");
@@ -214,30 +378,30 @@ static int test_cursor_beol_rtl (GtkHTML *html)
 	    || html->engine->cursor->offset != 35)
 		return FALSE;
 
-/* 	printf ("test_cursor_beol_rtl: testing 1st line bol\n"); */
-/* 	if (!html_engine_beginning_of_line (html->engine) */
-/* 	    || html->engine->cursor->offset != html->engine->cursor->position */
-/* 	    || html->engine->cursor->offset != 0) */
-/* 		return FALSE; */
+	printf ("test_cursor_beol_rtl: testing 1st line bol\n");
+	if (!html_engine_beginning_of_line (html->engine)
+	    || html->engine->cursor->offset != html->engine->cursor->position
+	    || html->engine->cursor->offset != 0)
+		return FALSE;
 
-/* 	/\* move to 2nd and test again *\/ */
-/* 	printf ("test_cursor_beol_rtl: testing 2nd line\n"); */
-/* 	if (!html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_DOWN, 1)) */
-/* 		return FALSE; */
+	/* move to 2nd and test again */
+	printf ("test_cursor_beol_rtl: testing 2nd line\n");
+	if (!html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_DOWN, 1))
+		return FALSE;
 
-/* 	printf ("test_cursor_beol_rtl: testing 2nd line eol\n"); */
-/* 	if (!html_engine_end_of_line (html->engine) */
-/* 	    || html->engine->cursor->offset != 11 */
-/* 	    || html->engine->cursor->position != 23) */
-/* 		return FALSE; */
+	printf ("test_cursor_beol_rtl: testing 2nd line eol\n");
+	if (!html_engine_end_of_line (html->engine)
+	    || html->engine->cursor->offset != 22
+	    || html->engine->cursor->position != 58)
+		return FALSE;
 
-/* 	printf ("test_cursor_beol_rtl: testing 2nd line bol\n"); */
-/* 	if (!html_engine_beginning_of_line (html->engine) */
-/* 	    || html->engine->cursor->offset != 0 */
-/* 	    || html->engine->cursor->position != 12) */
-/* 		return FALSE; */
+	printf ("test_cursor_beol_rtl: testing 2nd line bol\n");
+	if (!html_engine_beginning_of_line (html->engine)
+	    || html->engine->cursor->offset != 0
+	    || html->engine->cursor->position != 36)
+		return FALSE;
 
-/* 	printf ("test_cursor_beol_rtl: passed\n"); */
+	printf ("test_cursor_beol_rtl: passed\n");
 
 	return TRUE;
 }
@@ -270,6 +434,42 @@ test_quotes_in_div_block (GtkHTML *html)
 	return TRUE;
 }
 
+static int
+test_quotes_in_table (GtkHTML *html)
+{
+	GByteArray *flow_levels;
+	HTMLEngine *e = html->engine;
+	HTMLTable *table;
+
+	load_editable (html,
+		       "<blockquote type=cite>"
+		       "quoted text"
+		       "<table><tr><td>"
+		       "quoted text without inner cite"
+		       "</td></tr></table>"
+		       "quoted text"
+		       "</blockquote>");
+
+	/* test for right object hierarhy */
+	if (!e->clue || !HTML_CLUE (e->clue)->head || !HTML_CLUE (e->clue)->head->next || !HTML_IS_CLUEFLOW (HTML_CLUE (e->clue)->head->next)
+	    || !HTML_CLUE (HTML_CLUE (e->clue)->head->next)->head || !HTML_IS_TABLE (HTML_CLUE (HTML_CLUE (e->clue)->head->next)->head))
+		return FALSE;
+
+ 	table = HTML_TABLE (HTML_CLUE (HTML_CLUE (e->clue)->head->next)->head);
+
+ 	if (table->totalCols != 1 || table->totalRows != 1 || !table->cells [0][0] || !HTML_IS_TABLE_CELL (table->cells [0][0])
+	    || !HTML_CLUE (table->cells [0][0])->head || !HTML_IS_CLUEFLOW (HTML_CLUE (table->cells [0][0])->head))
+ 		return FALSE;
+
+	flow_levels = HTML_CLUEFLOW (HTML_CLUE (table->cells [0][0])->head)->levels;
+
+	/* test if levels are OK */
+	if (!flow_levels || !flow_levels->len == 0)
+		return FALSE;
+
+	return TRUE;
+}
+
 static int
 test_capitalize_upcase_lowcase_word (GtkHTML *html)
 {
@@ -347,12 +547,12 @@ int main (int argc, char *argv[])
 	gtk_html_set_editable (html, TRUE);
 	win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 	gtk_container_add (GTK_CONTAINER (win), html_widget);
-	gtk_widget_show_all (win);
-	gtk_widget_show_now (win);
+
+	/* gtk_widget_show_all (win); */
 
 	n_all = n_successful = 0;
 
-	fprintf (stderr, "GtkHTML test suite\n");
+	fprintf (stderr, "\nGtkHTML test suite\n");
 	fprintf (stderr, "--------------------------------------------------------------------------------\n");
 	for (i = 0; tests [i].name; i ++) {
 		int j, result;
@@ -373,7 +573,7 @@ int main (int argc, char *argv[])
 	}
 
 	fprintf (stderr, "--------------------------------------------------------------------------------\n");
-	fprintf (stderr, "%d tests failed %d tests passed (of %d tests)\n", n_all - n_successful, n_successful, n_all);
+	fprintf (stderr, "%d tests failed %d tests passed (of %d tests)\n\n", n_all - n_successful, n_successful, n_all);
 
 	return n_all != n_successful;
 }
